import unittest

import numpy as np

from run_batched import run_batched


def instrumental_keyed(x, batch_size_callback):
    assert isinstance(x, dict)
    assert x.keys() == {"a", "b"}
    batch_size_callback({k: len(v) for k, v in x.items()})
    return {"c": x["a"] + x["b"], "d": x["a"] - x["b"]}


def return_keyed_nested(x, batch_size_callback):
    batch_size_callback(len(x))
    return {"a": x, "b": {"c": x + 1}}


def input_keyed_nested(x):
    return x["a"]["b"] - x["c"]["d"]["e"]


class TestBasic(unittest.TestCase):
    def test_keyed_basic(self):
        batch_sizes = []
        res = run_batched(
            lambda x: instrumental_keyed(x, batch_sizes.append),
            {
                "a": np.arange(10),
                "b": np.arange(10) * 10,
            },
            2,
            device="cpu",
        )
        self.assertEqual(batch_sizes, [{"a": 2, "b": 2}] * 5)

        self.assertEqual(
            list(res.keys()),
            ["c", "d"],
        )
        res = {k: v.tolist() for k, v in res.items()}
        self.assertEqual(
            res,
            {
                "c": [0, 11, 22, 33, 44, 55, 66, 77, 88, 99],
                "d": [0, -9, -18, -27, -36, -45, -54, -63, -72, -81],
            },
        )

    def test_inconsistent_sizes(self):
        with self.assertRaisesRegex(
            ValueError, "All arrays must have the same length, but got: {2, 3}"
        ):
            run_batched(
                lambda x: instrumental_keyed(x, lambda _: None),
                {
                    "a": np.arange(2),
                    "b": np.arange(3),
                },
                2,
                device="cpu",
            )

    def test_return_keyed_nested(self):
        batch_sizes = []
        res = run_batched(
            lambda x: return_keyed_nested(x, batch_sizes.append),
            np.arange(10),
            2,
            device="cpu",
        )
        self.assertEqual(batch_sizes, [2] * 5)

        self.assertEqual(
            list(res.keys()),
            ["a", "b"],
        )
        self.assertEqual(
            list(res["b"].keys()),
            ["c"],
        )
        self.assertEqual(res["a"].tolist(), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        self.assertEqual(res["b"]["c"].tolist(), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

    def test_input_keyed_nested(self):
        res = run_batched(
            input_keyed_nested,
            {
                "a": {"b": np.arange(10)},
                "c": {"d": {"e": np.arange(10) * 3}},
            },
            2,
            device="cpu",
        )
        self.assertEqual(res.tolist(), [0, -2, -4, -6, -8, -10, -12, -14, -16, -18])



import numpy as np
import torch


def run_batched(m, x, bs, pbar=lambda x: x, *, device="cuda"):
    """
    Run m on x in batches of size bs, and concatenate the results.

    If x is a dict, treat it as a dict of arrays, each of which is batched
        together. This requires that the length of each array is the same.

    If m returns a dict, treat it as a dict of arrays, each of which is
        concatenated together. This requires that the length of each array is
        the same.

    Arguments
    ---------
    m : function
        A function which takes a batch of data and returns a batch of data.
    x : array or dict of arrays
        The data to be batched.
    bs : int
        The batch size.
    pbar : function
        A function which takes an iterable and returns an iterable. This is
        used to wrap the iterable over the batches of data to produce a pbar.

    Returns
    -------
    array or dict of arrays
        The result of running m on x in batches of size bs.
    """
    if not isinstance(bs, int):
        raise ValueError("Batch size must be an integer.")
    if bs <= 0:
        raise ValueError("Batch size must be positive.")
    size = overall_size(x)
    assert size != 0

    with torch.no_grad():
        ys = []
        for i in pbar(range((size + bs - 1) // bs)):
            x_to_use = select(x, i * bs, (i + 1) * bs, device)
            y = m(x_to_use)
            y = to_numpy(y)
            ys.append(y)
        result = concatenate_all(ys)
        return result


def overall_size(x):
    """
    Compute the overall size of an array or dict of arrays.

    Arguments
    ---------
    x : array or dict of arrays
        The arrays to be measured.

    Returns
    -------
    int
    """
    if isinstance(x, dict):
        each = {overall_size(v) for v in x.values()}
        if len(each) != 1:
            raise ValueError(f"All arrays must have the same length, but got: {each}")
        return each.pop()
    return x.shape[0]


def select(x, start, end, device):
    """
    Select a range of indices from an array or dict of arrays.

    Arguments
    ---------
    x : array or dict of arrays
        The arrays to be sliced.
    start : int
        The start index.
    end : int
        The end index.

    Returns
    -------
    array or dict of arrays
    """
    if isinstance(x, dict):
        return {k: select(v, start, end, device) for k, v in x.items()}
    return torch.tensor(x[start:end]).to(device)


def concatenate_all(ys):
    """
    Concatenate a list of arrays or dicts of arrays.

    Arguments
    ---------
    ys : list of arrays or dicts of arrays
        The arrays to be concatenated.

    Returns
    -------
    array or dict of arrays
    """
    if isinstance(ys[0], dict):
        assert all(y.keys() == ys[0].keys() for y in ys)
        result = {k: concatenate_all([y[k] for y in ys]) for k in ys[0]}
    else:
        result = np.concatenate(ys)
    return result


def to_numpy(y):
    """
    Convert the following tensor or dict of tensors to numpy arrays.
    """
    if isinstance(y, dict):
        y = {k: to_numpy(y[k]) for k in y}
    else:
        y = y.cpu().numpy()
    return y



import unittest

import numpy as np
import torch

from run_batched import run_batched


def instrumented_double(x, batch_size_callback):
    assert isinstance(x, torch.Tensor)
    batch_size_callback(len(x))
    return 2 * x


def subtract_mean(x, batch_size_callback):
    assert isinstance(x, torch.Tensor)
    batch_size_callback(len(x))
    return x - torch.mean(x)


class TestBasic(unittest.TestCase):
    def test_addition(self):
        batch_sizes = []
        self.assertEqual(
            run_batched(
                lambda x: instrumented_double(x, batch_sizes.append),
                np.array([1, 2, 3]),
                2,
                device="cpu",
            ).tolist(),
            [2, 4, 6],
        )
        self.assertEqual(batch_sizes, [2, 1])

    def test_extremely_large_batch(self):
        batch_sizes = []
        self.assertEqual(
            run_batched(
                lambda x: instrumented_double(x, batch_sizes.append),
                np.array([1, 2, 3, 4, 5, 6]),
                100,
                device="cpu",
            ).tolist(),
            [2, 4, 6, 8, 10, 12],
        )
        self.assertEqual(batch_sizes, [6])

    def test_invalid_batch_size_not_int(self):
        with self.assertRaisesRegex(ValueError, "Batch size must be an integer."):
            run_batched(lambda x: x, np.array([1, 2, 3]), 2.5, device="cpu")

    def test_invalid_batch_size_negative(self):
        with self.assertRaisesRegex(ValueError, "Batch size must be positive."):
            run_batched(lambda x: x, np.array([1, 2, 3]), -1, device="cpu")

    def test_only_batch_first_axis(self):
        x = np.arange(2400).reshape(40, 3, -1)
        batch_sizes = []
        self.assertEqual(
            run_batched(
                lambda x: instrumented_double(x, batch_sizes.append),
                x,
                2,
                device="cpu",
            ).tolist(),
            (2 * x).tolist(),
        )
        self.assertEqual(batch_sizes, [2] * 20)

    def test_batches_consecutive(self):
        x = np.arange(10, dtype=np.float32)
        batch_sizes = []
        self.assertEqual(
            run_batched(
                lambda x: subtract_mean(x, batch_sizes.append),
                x,
                3,
                device="cpu",
            ).tolist(),
            [-1, 0, 1, -1, 0, 1, -1, 0, 1, 0],
        )
        self.assertEqual(batch_sizes, [3, 3, 3, 1])



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="run_batched",
    version="1.0.0",
    author="Kavi Gupta",
    author_email="run-batched@kavigupta.org",
    description="Permanant cache.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kavigupta/run_batched",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
    install_requires=["torch", "numpy"],
)
